/**
 * @author zAlbee
 * Created on May 3, 2005
 */
import java.io.*;
import java.util.*;
import java.text.*;
import java.nio.charset.Charset;
import com.freeware.inifiles.INIFile;

public class immerge {
	
	static final String VERSION     = "0.66";
	static final String LAST_MODIFY = "Sep 22, 2009";

	static final String NL = System.getProperty("line.separator"); // newline (CR/CRLF/LF)

	// The Byte-Order Mark (BOM), when occuring at the beginning of a file, indicates a Unicode file 
	// It is the single Unicode character U+FEFF (ZERO WIDTH NO-BREAK SPACE)
	// Generally, it is a convention of Win32 programs to place the BOM, but not POSIX systems.
	static final char UNICODEBOM = 0xFEFF; // BOM as Unicode character
	static final byte[] UTF8BOM = {(byte)0xEF, (byte)0xBB, (byte)0xBF}; // BOM encoded in UTF-8

	static final String INI_SECTION_ICQNAMES = "ICQ Names";
	static final String INI_SECTION_MSNUSERS = "MSN UserIDs";
	static final int AIM = 0;
	static final int ICQ = 1;
	static final int IRC = 2;
	static final int MSN = 3;
	static final int YAHOO = 4;
	static final int ASTRA = 5;
	static final int FACEBOOK = 6;
	static final int GOOGLE = 7;
	static final int JABBER = 8;
	static final int MYSPACE = 9;
	static final int XFIRE = 10;
	
	static int tz_adjust = -16;
	
	static void debug(String s) {
		//System.out.println(s);
	}
	
	static INIFile settingsINI;
	
	static boolean promptOnNewName;
	static boolean promptOnVagueName;
	
	static HashSet yourDisplayNames = new HashSet();
	static HashSet userDisplayNames = new HashSet();
	
	static String defaultCharsetName;
	
	static BufferedReader in;
	
	/***
	 * Log abstract class, associated with a BufferedReader
	 *
	 * Subclasses need to implement:
	 *
	 *   readSession() - reads the next session into a buffer
	 *              - puts that session's date into the sessDate property.
	 *              - returns True if a session was read, False if the log was empty (no session read)
	 *
	 *   getSession() - should return all the text within that session.
	 */
	static abstract class Log implements Comparable {
		static final String sessStart  = "Session Start (";
		static final String sessStart2 = "sion Start (";
		static final String sessClose  = "Session Close (";
		static SimpleDateFormat sdfTrillian = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy", Locale.US);
		static SimpleDateFormat sdfTrillianTZ = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy Z", Locale.US);
		static SimpleDateFormat sdfICQ = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss", Locale.US);
		static SimpleDateFormat sdfTTime = new SimpleDateFormat("[hh:mm a]", Locale.US);
		static SimpleDateFormat sdfTTime24 = new SimpleDateFormat("[HH:mm]", Locale.US);
		static SimpleDateFormat sdfMSN = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'", Locale.US);
		
		String trillID; // hack to save name before lowercasing (v0.63)
		
		String filename;
		String pathname;
		MyReader reader;
		Date sessDate;
		String buffer;
		boolean empty = false;
		int priority = 0; // when multiple logs have the same session date, the higher priority log's session is used.

		// 0.61
		File file;
		String charsetName;

		abstract public boolean readSession() throws Exception;
		abstract public String getSession();

		/** Log() constructor - automatic UTF-8 detection
		 */
		Log(File file) throws IOException {
			this(file, null);
		}

		/** Log() constructor - explicit charset
		 */
		Log(File file, String charsetName) throws IOException {
			filename = file.getName();
			pathname = file.getPath();
			this.file = file;
			this.charsetName = charsetName;
		}
		
		void init() throws IOException {
			init(false);
		}
		void init(boolean forceInit) throws IOException {
			if (!forceInit && reader != null) return;
			
			// forceInit == true OR reader == null
			if (charsetName == null) {
				// every Java VM supports ISO-8859-1
				reader = new MyReader(new InputStreamReader(new FileInputStream(file), "iso-8859-1"));
				
				reader.mark(16);
				if (reader.read() == 0xEF && reader.read() == 0xBB && reader.read() == 0xBF) { 
					// Detected UTF-8 header:
					// Use UTF-8 reader, but we still need to ignore the first 3 bytes (1 character).
					
					InputStreamReader in = new InputStreamReader(new FileInputStream(file), "utf-8");
					reader = new MyReader(in);
					
					reader.mark(16);
					if (reader.read() == 0xFEFF) {
					}
					else {
						// As long as Java's UTF-8 reader works, this should never happen
						// Maybe if the system is byte-swapped.
						System.out.println("Warning: UTF-8 BOM was detected, but was not treated as such. Send a bug report to zalbee" + "@gmail.com");
						reader.reset();
					}
				}
				//else reader.reset();
				else {
					if (defaultCharsetName == null) reader = new MyReader(new FileReader(file));
					else reader = new MyReader(new InputStreamReader(new FileInputStream(file), defaultCharsetName));
				}
			}
			
			else {
				InputStreamReader in = new InputStreamReader(new FileInputStream(file), charsetName);
				reader = new MyReader(in);
					
				reader.mark(16);
				if (in.getEncoding() == "UTF8" && reader.read() == 0xFEFF) {
				}
				else {
					reader.reset();
				}
			}

		}
		
		/** Allows sorting instances of Log by date and priority.
		 */
		public int compareTo(Object o2) {
			Log log2 = (Log)o2;
			int cmp = sessDate.compareTo(log2.sessDate);
			if (cmp == 0) return log2.priority - priority;
			else return cmp;
		}

	} // end abstract class Log
	
	
	/***
	 * ConverterLog abstract class, subclass of Log
	 *
	 *   This is for log formats that store history in a different way than Trillian.
	 *   This class has methods for converting single messages to Trillian-style messages and sessions.
	 *   Most non-Trillian log formats (eg. ICQ, MSN) will use this class - TrLog and TrXML do NOT
	 *   inherit this class because they directly copy and do not convert the messages.
	 *
	 * Defines methods:
	 *
	 *   readSession() - Reads the next session into buffer. A session is multiple messages within a certain timeframe.
	 *   getTrMsg()     - Returns a Trillian-log formatted version of the stored message.
	 *   getTrSession() - Returns a Trillian-log formatted version of the stored session.
	 *   getXMLMsg()     - Returns a Trillian-XML formatted version of the stored message.
	 *   getXMLSession() - Returns a Trillian-XML formatted version of the stored session.
	 * 
	 * Subclasses need to implement:
	 *
	 *   readMsg() - reads the next message (as plaintext, with no entities, and no extra meta-information) into msgBuffer
	 *             - msgBuffer should not have an extra newline character at the end (unless that is part of the message)
	 *             - returns true if a message was read, otherwise false
	 *             - sets msgDate to the date of the message
	 *             - sets msgIsSend = true if that message is sent from the owner of this account 
	 *               (yourNick), otherwise false
	 *             - can also set properties yourNick and nickName
	 *
	 *   getMsg() - suggested to point to one of getTrMsg() or getXMLMsg()
	 *
	 *   getSession() - suggested to point to one of getTrSession() or getXMLSession()
	 *
	 *   Note: If you plan to override the readSession() method, then it is not necessary to follow the above guidelines
	 *   for readMsg(), getMsg(), and getSession().
	 *
	 */
	static abstract class ConverterLog extends Log {
		protected String msgBuffer;
		protected Date msgDate;
		protected boolean msgIsSend;

		protected String msgBuffer2;
		protected Date msgBuffer2Date = null;

		protected Date sessEndDate;
		
		// It is up to the subclass to set these:
		protected String medium;
		protected String userID;
		protected String yourUserID;
		protected String nickName;
		protected String yourNick;
		
		private String encUserID;
		private String encYourUserID;
		private String encNickName;
		private String encYourNick;
		
		public void setUserID(String theirUIN) {
			this.userID = theirUIN;
			encUserID = trXMLEntity(userID);
		}
		
		public void setYourUserID(String yourUIN) {
			this.yourUserID = yourUIN;
			encYourUserID = trXMLEntity(yourUserID);
		}

		public void setNick(String theirNick) {
			this.nickName = theirNick;
			encNickName = trXMLEntity(nickName);
		}

		public void setYourNick(String yourNick) {
			this.yourNick = yourNick;
			encYourNick = trXMLEntity(yourNick);
		}

		public String getUserID() {
			return userID;
		}

		public String getNick() {
			return nickName;
		}

		public String getYourNick() {
			return yourNick;
		}

		public String getYourUserID() {
			return yourUserID;
		}

		public ConverterLog(File file) throws IOException {
			super(file);
		}

		public ConverterLog(File file, String charsetName) throws IOException {
			super(file, charsetName);
		}
		
		abstract protected boolean readMsg() throws Exception;
		abstract protected String getMsg();

		/** ConverterLog.readSession()
		 *   Reads one session and stores it.
		 *
		 *   This method calls readMsg() and getMsg() multiple times, and concatenates the returned Strings into one
		 *   session String. The number of messages in a single session equals the number of messages
		 *   that are dated within 24 hours of the first message in that session.
		 */
		public boolean readSession() throws Exception {
			sessDate = msgBuffer2Date;
			String msg = null;

			if (sessDate == null) {
				boolean read = readMsg();
				if (!read) {empty=true; reader.close(); return false;}

				msg = getMsg();
				sessDate = msgDate;
			}
			else {
				msg = msgBuffer2;
				msgBuffer2 = null;
				msgBuffer2Date = null;
			}

			sessEndDate = sessDate;
			buffer = msg;

			boolean hasMsg = readMsg();

			while (hasMsg) {
				msg = getMsg();

				if (msgDate.getTime() < sessDate.getTime() + 24*60*60*1000) {
					buffer += msg;
					sessEndDate = msgDate;
				}
				else {
					msgBuffer2 = msg;
					msgBuffer2Date = msgDate;
					break;
				}

				hasMsg = readMsg();
			}

			return true;
		}

		/** ConverterLog.getTrMsg()
		 *   Returns a Trillian-log formatted version of the stored message.
		 */
		protected String getTrMsg() {
			String prefix = sdfTTime.format(msgDate) + " ";
			if (msgIsSend) prefix += yourNick + ": ";
			else prefix += nickName + ": ";
			
			String suffix = NL;

			String tmpMsg = msgBuffer.replaceAll("\n", "\n ");
			
			return prefix + tmpMsg + suffix;
		}

		/** ConverterLog.getTrSession()
		 *   Returns a Trillian-log formatted version of the stored session.
		 */
		protected String getTrSession() {
			String prefix = sessStart + (yourUserID.length()>0 ? yourUserID : medium) + ":" + nickName + "): " + sdfTrillianTZ.format(sessDate) + NL;
			String suffix = sessClose + nickName + "): " + sdfTrillianTZ.format(sessEndDate) + NL+NL+NL;

			return prefix + buffer + suffix;
		}

		/** ConverterLog.getXMLMsg()
		 *   Returns a Trillian-XML formatted version of the stored message.
		 */
		protected String getXMLMsg() {
			//<message type="incoming_privateMessage" time="1132707630" medium="MSN" to="zalbee2%40hotmail%2Ecom" from="catty_kitty_leung%40hotmail%2Ecom" from_display="%28%40%29%20SuPeR%20cAt%20%2D%20Jay%20Chow%20is%20amazing" text="that%20was%20a%20FAST%20response"/>
			//<message type="outgoing_privateMessage" time="1133042238" medium="ICQ" to="16225800" from="18196095" from_display="zAlbee" text="hm%2E"/>
			String prefix = "<message type=\"" + (msgIsSend ? "outgoing_privateMessage" : "incoming_privateMessage")
					+ "\" time=\"" + msgDate.getTime()/1000 + "\" medium=\"" + medium + "\" to=\""
					+ (msgIsSend ? encUserID : encYourUserID) + "\" from=\"" + (msgIsSend ? encYourUserID : encUserID)
					+ "\" from_display=\"" + (msgIsSend ? encYourNick : encNickName) + "\" text=\"";
			
			String suffix = "\"/>" + NL;
			
			//String tmpMsg = msgBuffer.replaceAll("\n", "\n ").replaceAll("&","&amp;").replaceAll("<","&lt;");
			String tmpMsg = msgBuffer.replaceAll("\n", "\n ");

			return prefix + trXMLEntity(tmpMsg) + suffix;
		}

		/** ConverterLog.getXMLSession()
		 *   Returns a Trillian-XML formatted version of the stored session.
		 */
		protected String getXMLSession() {
			//<session type="start" time="1133041383" medium="ICQ" to="16225800" from="18196095"/>
			String prefix = "<session type=\"start\" time=\"" + sessDate.getTime()/1000 +  "\" medium=\"" + medium + "\" to=\"" 
					+ encUserID +"\" from=\""+ encYourUserID + "\"/>" + NL;

			//<session type="stop" time="1133060545" medium="ICQ" to="16225800" from="18196095"/>
			String suffix = "<session type=\"stop\" time=\"" + sessEndDate.getTime()/1000 +  "\" medium=\"" + medium + "\" to=\"" 
					+ encUserID +"\" from=\""+ encYourUserID + "\"/>" + NL;
					
			return prefix + buffer + suffix;
		}
	} // end abstract class ConverterLog


	/***
	 * ICQHRLog class, subclass of ConverterLog
	 * Represents an ICQ log file (ICQ History Reader HTML format)
	 */
	static class ICQHRLog extends ConverterLog {
		private String sendColour;
		private String recvColour;
		private static final String genStart = "<font size=3 color=";
		private static final String genStart2 = ">";
		private static final String genEnd = "</font>";
		private static final int START_LEN = genStart.length() + 8;
		private String sendStart;
		private String recvStart;
		
		/** ICQHRLog() contructor
		 */
		public ICQHRLog(File f) throws IOException {
			super(f);
			priority = 2;
			medium = "ICQ";
		}
		
		/** ICQHRLog.init()
		 * Sets userID, yourUserID, yourNick. Tries to find nickName from log.
		 * Prepares log for reading.
		 */
		
		public void init() throws IOException {
			init(false);
		}

		public void init(boolean forceInit) throws IOException {
			if (!forceInit && reader != null) return;

			super.init(forceInit);

			String s = reader.readLine();
			
			while (s != null) {
				if (s.matches("<td [^>]*>Nickname:</td>")) {
					s=reader.readLine();
					setNick( s.substring( s.indexOf('>') +1, s.lastIndexOf("</td>") ) );
					break;
				}
				else if (s.trim().toLowerCase().equals("no info for this user")) {
					System.out.println("Info: No nickname in " + filename);
					setNick(userID);
					s = reader.readLine();
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Reached end-of-file while looking for info in " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Sent</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					sendColour = s.substring( colIdx, colIdx + 7);
					sendStart = genStart + sendColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read sent colour from " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Recieved</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					recvColour = s.substring( colIdx, colIdx + 7);
					recvStart = genStart + recvColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read receive colour from " + filename);
				empty = true;
				return;
			}

		}

		/** ICQHRLog.readMsg()
		 *    returns true if a message was read, otherwise false
		 *    use getMsg() to see the results of this operation
		 */
		protected boolean readMsg() throws Exception {
			boolean msgOpen = false;
			msgBuffer = "";

			String s=reader.readLine();
			msgDate = null;

			// grab date

			while (s != null && msgDate==null) {
				if ( s.startsWith(genStart) ) {
					int dateStartIdx = s.indexOf('(', START_LEN) + 1;
					int dateEndIdx = s.lastIndexOf(')');
					msgDate = sdfICQ.parse(s.substring(dateStartIdx, dateEndIdx));
					//dpoint[0].setTime( dpoint[0].getTime() - 16*60*60*1000 ); // has sync problems when dates are near daylight savings switch
					msgDate.setHours( msgDate.getHours() + tz_adjust); // -16 hours seems to make the time correct? no idea how ICQHR calculates timezone
				}
				s = reader.readLine();
			}
			
			// grab colour (to determine sender) and message

			while (s != null && !msgOpen) {
				if ( s.startsWith(genStart)) {
					msgOpen = true;
					msgIsSend = s.startsWith(sendStart);

					if (s.trim().endsWith(genEnd)) {
						msgBuffer += unHTMLEntity( s.substring( START_LEN, s.lastIndexOf(genEnd) ) );
						msgOpen = false;
						return true;
					}//break;}
					else if (s.trim().endsWith("<BR>")) msgBuffer += unHTMLEntity( s.substring( START_LEN, s.lastIndexOf("<BR>") ) ) + NL;
					else msgBuffer += unHTMLEntity( s.substring( START_LEN ) ) + NL;
				}

				s=reader.readLine();
			}

			// if message consists of more than one line

			while (s != null && msgOpen) {
				if (s.trim().endsWith(genEnd)) {
					msgBuffer += unHTMLEntity( s.substring( 0, s.lastIndexOf(genEnd)) );
					msgOpen = false;
					return true;
				} //break;}
				else if (s.trim().endsWith("<BR>")) msgBuffer += unHTMLEntity( s.substring( 0, s.lastIndexOf("<BR>") ) ) + NL;
				else msgBuffer += unHTMLEntity(s) + NL;

				s=reader.readLine();
			}
			
			// if we're here, message didn't terminate, OR there was no message in the first palce
			
			if (msgOpen) return true;
			else return false;

			//return msg;
		}
		
		/** ICQHRLog.getMsg()
		 */
		protected String getMsg() {
			return getTrMsg();
		}

		/** ICQHRLog.getSession()
		 */
		public String getSession() {
			return getTrSession();
		}

	} //end class ICQHRLog
	
	
	/***
	 * ICQLiteLog class, subclass of ConverterLog
	 * Represents an ICQ Lite log file
	 */
	static class ICQLiteLog extends ConverterLog {
		
		/** ICQLiteLog() contructor
		 */
		public ICQLiteLog(File f) throws IOException {
			super(f, "utf-8");
			priority = 2;
			medium = "ICQ";
		}

		/** ICQLiteLog.readMsg()
		 *    returns true if a message was read, otherwise false
		 *    use getMsg() to see the results of this operation
		 */
		protected boolean readMsg() throws Exception {
			boolean eventOpen = false;
			boolean hasMsg = false;
			
			msgBuffer = "";

			String s = reader.readToNext(">");
			msgDate = null;
			
			// parse XML. with XML, tags can occur in any order.
			
			while (s != null) {
				String str = s.trim();
				
				if (!eventOpen) {
					if (str.equals("<event>")) {
						eventOpen = true;
					}
				}
				else {
					if (str.equals("<text>")) {
						msgBuffer = unHTMLEntity2(reader.readToNext("</text>", false));
						if (msgBuffer.endsWith("\r\n")) {
							msgBuffer = msgBuffer.substring(0, msgBuffer.length() - 2);
						}
						else if (msgBuffer.endsWith("\n") || msgBuffer.endsWith("\r")) {
							msgBuffer = msgBuffer.substring(0, msgBuffer.length() - 1);
						}
						hasMsg = true;
					}
					else if (str.equals("</event>")) {
						eventOpen = false;
						// must break here, to prevent extra read
						if (hasMsg) break;
					}
	
					else if (str.equals("<incoming>")) {
						if (reader.readToNext("</incoming>",false).equals("Yes")) {
							msgIsSend = false;
						}
						else {
							msgIsSend = true;
						}
					}
					else if (str.equals("<time>")) {
						String secs = reader.readToNext("</time>",false);
						msgDate = new Date(Long.parseLong(secs) * 1000);
					}
				} // end if eventOpen

				s = reader.readToNext(">");
			} // end while s
			
			return hasMsg;
		}
		
		/** ICQLiteLog.getMsg()
		 */
		protected String getMsg() {
			return getTrMsg();
		}

		/** ICQLiteLog.getSession()
		 */
		public String getSession() {
			return getTrSession();
		}

	} //end class ICQLiteLog


	/***
	 * ICQHRtoXML class, subclass of ICQHRLog
	 * Represents the XML version of the ICQ History Reader file
	 */
	static class ICQHRtoXML extends ICQHRLog {
		/** ICQHRtoXML() contructor
		 */
		public ICQHRtoXML(File f) throws IOException {
			super(f);
			priority = 2;
		}
		
		/** ICQHRLog.getMsg()
		 */
		protected String getMsg() {
			return getXMLMsg();
		}

		/** ICQHRLog.getSession()
		 */
		public String getSession() {
			return getXMLSession();
		}
	}


	/***
	 * ICQLitetoXML class, subclass of ICQLiteLog
	 * Represents the XML version of the ICQ Lite log file
	 */
	static class ICQLitetoXML extends ICQLiteLog {
		/** ICQLitetoXML() contructor
		 */
		public ICQLitetoXML(File f) throws IOException {
			super(f);
			priority = 2;
		}
		
		/** ICQLiteLog.getMsg()
		 */
		protected String getMsg() {
			return getXMLMsg();
		}

		/** ICQLiteLog.getSession()
		 */
		public String getSession() {
			return getXMLSession();
		}
	}


	/***
	 * TrLog class, subclass of Log
	 * Represents a Trillian logfile.
	 */
	static class TrLog extends Log {
		private boolean sessOpen = false;
		private boolean startOfFile = true;

		/** TrLog() contructor
		 */
		public TrLog(File f) throws IOException {
			super(f);
		}

		/** TrLog.readSession()
		 */
		public boolean readSession() throws Exception {
			sessOpen = false;
			String s = reader.readLine();
			StringBuffer sb = new StringBuffer();

			int nameEndIdx;

			while (s != null) {
				if (!sessOpen) {
					if (    ( s.startsWith( sessStart ) || (startOfFile && s.startsWith(sessStart2)) )
						 && ( nameEndIdx=s.lastIndexOf("):", s.length()-24) ) > -1
						) {
						int dateStartIdx = nameEndIdx + 3;
	
						try {
							sessDate = sdfTrillianTZ.parse(s.substring(dateStartIdx));
						}
						catch (ParseException e) {
							sessDate = sdfTrillian.parse(s.substring(dateStartIdx));
						}
	
						if (s.startsWith("sion")) {
							//s = "Ses" + s;
							sb.append("Ses");
						}
						sessOpen = true;
						//buffer = s + NL;
						sb.append(s).append(NL);
					}
					else if (s.trim().length()>0) {
						// Text exists outside a session. Write it anyway to keep fidelity of logs.
						// Give it a "dummy" session, with the same date as the last one, so that it's written right away.
						// Since there will be no session close marker, Read until the next session starts, then rewind the reader.
						
						//System.out.println("Warning: Including text out of session = " + s);
						sessOpen = true;
						//buffer = s + NL;
						sb.append(s).append(NL);
					}
				} // end if !sessOpen
				
				else { // sessOpen

					if ( s.startsWith(sessStart) && ( nameEndIdx=s.lastIndexOf("):", s.length()-24) ) > -1 ) {
						// Encountered a new session. Rewind to before this line.
						reader.reset();
						sessOpen = false;
						//if (!buffer.endsWith(NL+NL)) {
						if (!sb.toString().endsWith(NL+NL)) {
							//buffer += NL + NL; // insert some blank lines to make it easier on the eyes
							sb.append(NL).append(NL);
						}
						break;
					}

					//buffer += s + NL;
					sb.append(s).append(NL);

					if ( s.startsWith( sessClose ) && ( nameEndIdx=s.lastIndexOf("):", s.length()-24) ) > -1 ) {
						// Found session end.
				  		sessOpen = false;
				  		//buffer += NL + NL;
						sb.append(NL).append(NL);
				  		break;
					}

				} // end if sessOpen
				
				startOfFile = false;
				// in case have to rewind.
				/*
				Email address max == 64 + '@' + 255 <RFC 2822>
					  		  	  == 320
				MSN name max == 129;
				"Session Start (email@domain.com:LONG DISPLAY NAME): Sun Jul 25 00:54:39 2004 -0500"
				Total == 15 + 320 + ":" + 129 + 27 + 5
					  == 497
				*/
				reader.mark(512);
				s = reader.readLine();

			} // end while s

			if (s==null) {
				// Reached EOF
				if (sessOpen) ; // not terminated correctly. write it anyway.
				else {
					empty = true;
					reader.close();
					return false;
				}
			}
			buffer = sb.toString();
			return true;

		}

		/** TrLog.getSession()
		 */
		public String getSession() {
			return buffer;
		}
	} //end class TrLog


	/***
	 * TrXML class, subclass of Log
	 * Represents a Trillian XML log file (in the "query" folder)
	 */
	static class TrXML extends Log {
		private boolean sessOpen = false;
		
		/** TrXML() contructor
		 */
		public TrXML(File f) throws IOException {
			super(f, "iso-8859-1");
			priority = 1;
		}

		public boolean readSession() throws Exception {
			sessOpen = false;
			String s = reader.readLine();
			StringBuffer sb = new StringBuffer();

			int nameEndIdx;

			while (s != null && !sessOpen) {
				if ( s.toLowerCase().startsWith( "<session type=\"start\"" ) ) {
					
					int dateStartIdx = s.indexOf("time=\"") + 6;
					int dateEndIdx = s.indexOf("\"", dateStartIdx);
					
					sessDate = new Date( 1000 * Long.parseLong( s.substring(dateStartIdx, dateEndIdx) ) );
					
					sessOpen = true;
					//buffer = s + NL;
					sb.append(s).append(NL);
				}
				else if (s.trim().length()>0) {
					//System.out.println("Warning: skipping line: " + s);
					//buffer = s + NL;
					sb.append(s).append(NL);
					reader.mark(1024);
					s = reader.readLine();
					
					while (s != null) {
						if ( s.toLowerCase().startsWith( "<session type=\"start\"" ) ) {
							reader.reset();
							break;
						}
						//buffer += s + NL;
						sb.append(s).append(NL);
						reader.mark(1024);
						s = reader.readLine();
					}
					buffer = sb.toString();
					return true;
				}
				
				s = reader.readLine();
			}

			while (s != null && sessOpen) {
				//buffer += s + NL;
				sb.append(s).append(NL);

				
				if ( s.toLowerCase().startsWith( "<session type=\"stop\"" ) ) {
			  		sessOpen = false;
			  		break;
				}
				s = reader.readLine();
			}

			if (s==null) {
				if (sessOpen) ; // not terminated correctly. write it anyway.
				else {
					empty = true;
					reader.close();
					return false;
				}
			}
			buffer = sb.toString();
			return true;
			
		}
		
		public String getSession() {
			return buffer;
		}
		
	} //end class TrXML

	/***
	 * MSNLog class, subclass of ConverterLog
	 * Represents an MSN/Windows Live Messenger log file
	 */
	static class MSNLog extends ConverterLog {
		
		private int currSess = 0;
		private int msgSessNum = 0;
		private String toName;
		private String fromName;
		protected String sessNick;
		protected boolean isGroupChat = false; // should be property of a session, but will be set by readMsg()
		protected String prevNick; // save the last known nickname
		protected String prevYourNick; // save the last known owner nickname


		/** MSNLog() contructor
		 */
		public MSNLog(File f) throws IOException {
			super(f, "utf8");
			priority = 2;
			medium = "MSN";
		}

		/**
		 * @override
		 */
		public void setNick(String theirNick) {
			setNick(theirNick, true);
		}
		public void setNick(String theirNick, boolean isReliable) {
			super.setNick(theirNick);
			if (isReliable) prevNick = theirNick;
		}

		/**
		 * @override
		 */
		public void setYourNick(String yourNick) {
			setYourNick(yourNick, true);
		}
		public void setYourNick(String yourNick, boolean isReliable) {
			super.setYourNick(yourNick);
			if (isReliable) prevYourNick = yourNick;
		}
		
		/** MSNLog.readSession() - nice version, with readMsg(currSess) param
		 *    - not usable because of mark
		 */
		public boolean readSession1() throws Exception {			
			boolean hasMsg = readMsg();
			
			if (!hasMsg) {
				empty = true;
				reader.close();
				return false;
			}
			buffer = "";
			sessDate = msgDate;
			sessEndDate = sessDate;

			while (hasMsg) {
				buffer += getMsg();
				sessEndDate = msgDate;
				//hasMsg = readMsg(currSess);
			}

			return true;
		}

		/** MSNLog.readSession()
		 *    overrides ConverterLog
		 */
		public boolean readSession() throws Exception {			
			String msg = null;
		
			if (msgBuffer2Date == null) {
				boolean read = readMsg();
				if (!read) {
					empty=true; reader.close(); return false;
				}
				msg = getMsg();
				sessDate = msgDate;
			}
			else {
				msg = msgBuffer2;
				sessDate = msgBuffer2Date;
				msgBuffer2 = null;
				msgBuffer2Date = null;
			}
			currSess = msgSessNum;
	
			sessEndDate = sessDate;
			buffer = msg;

			// nickname for session = first nickname, unless it is a group conversation
			sessNick = isGroupChat ? "Group Conversation" : nickName;
			
			boolean hasMsg = readMsg();

			while (hasMsg) {
				msg = getMsg();

				if (currSess == msgSessNum) {
					buffer += msg;
					sessEndDate = msgDate;
				}
				else {
					msgBuffer2 = msg;
					msgBuffer2Date = msgDate;
					break;
				}

				hasMsg = readMsg();
			}
	
			return true;
		}
		
		/** MSNLog.readMsg()
		 *    returns true if a message was read, otherwise false
		 *    use getMsg() to see the results of this operation
		 *
		 *    Sets msgSessNum, isGroupChat
		 *    Sets nickName, yourNick
		 *    Sets prevNick, prevYourNick
		 */
		protected boolean readMsg() throws Exception {
			boolean messageOpen = false;
			
			//reader.mark(1024); // 400 char max message length + extra
			String s = reader.readToNext(">");
			
			int startIdx;
			int endIdx;
			int nameEndIdx;
			
			while (s != null) {
				s = s.trim();
				if (!messageOpen) {
					if (s.startsWith("<Message ")) {
						messageOpen = true;
	
						startIdx = s.indexOf("DateTime=\"") + 10;
						endIdx = s.indexOf("\"", startIdx);
						
						msgDate = sdfMSN.parse( s.substring(startIdx, endIdx) );
						
						startIdx = s.indexOf("SessionID=\"") + 11;
						endIdx = s.indexOf("\"", startIdx);
						
						msgSessNum = Integer.parseInt( s.substring(startIdx, endIdx) );
					}
				}
				else if (s.startsWith("<From")) {
					s = reader.readToNext("</From>");
					
					startIdx = s.indexOf("FriendlyName=\"") + 14;
					endIdx = s.indexOf("\"", startIdx);
					
					fromName = unHTMLEntity2(s.substring(startIdx, endIdx));
				}
				else if (s.startsWith("<To")) {
					// Note: If this is a group conversation, there can be multiple <User>s in <To>
					s = reader.readToNext("</To>");
					
					startIdx = s.indexOf("FriendlyName=\"") + 14;
					endIdx = s.indexOf("\"", startIdx);

					toName = unHTMLEntity2(s.substring(startIdx, endIdx));
					
					if (s.indexOf("FriendlyName=\"", endIdx) !=-1) {
						isGroupChat = true;
						// to properly output in MSN native format, we'd have to support multiple toNames
						// ignore them for now
					}
					else {
						isGroupChat = false;
					}
				}
				else if (s.startsWith("<Text")) {
					s = reader.readToNext("</Text>", false);
					
					msgBuffer = unHTMLEntityMSN(s);
				}
				else if (s.equals("</Message>")) {
					// message complete
					// set nicknames

					boolean confirmedNew = false;

					// prevNick and prevYourNick are known to be correct from before.
					if      (prevYourNick.equals(fromName) || (!isGroupChat && prevNick.equals(toName)) ) msgIsSend = true;
					else if (prevYourNick.equals(toName)   || prevNick.equals(fromName)) msgIsSend = false;
					else if (yourDisplayNames.contains(fromName) || (!isGroupChat && userDisplayNames.contains(toName)) ) msgIsSend = true;
					else if (yourDisplayNames.contains(toName)   || userDisplayNames.contains(fromName)) msgIsSend = false;
					else {
						// do some probabilistic guess-work here
						float a = lcsP(fromName, prevYourNick);
						float b = lcsP(fromName, prevNick);
						float c = isGroupChat ? 0 : lcsP(toName, prevYourNick);
						float d = isGroupChat ? 0 : lcsP(toName, prevNick);
						
						// a,d mean message is outgoing, msgIsSend = true
						// b,c mean message is incoming, msgIsSend = false
						
						float diff = a+d -b-c;
						
						if (diff > 0) msgIsSend = true;
						else msgIsSend = false;
					
						System.out.println("Names changed: ");
						if (isGroupChat) {
							System.out.println(" IMmerge thinks " + fromName + (msgIsSend ? " = " : " is NOT ") + yourUserID);
						}
						else {
							System.out.println(" IMmerge thinks " + (msgIsSend ? fromName : toName ) + " = " + yourUserID);
							System.out.println(" IMmerge thinks " + (msgIsSend ? toName : fromName ) + " = " + userID );
						}
						System.out.println(" " + Math.abs(diff) + " certainty. ");
						
						if (promptOnNewName || (promptOnVagueName && Math.abs(diff) < 1.01 )) {
							System.out.print("Is this correct? (Yes/No) ");

							if (!readYesNoFrom(in)) {
								msgIsSend = !msgIsSend;
							}
							
							confirmedNew = true;
						}
					}

					if (msgIsSend) {
// 						yourNick = fromName;
// 						nickName = toName; // ***
// 						if (isGroupChat) prevYourNick = yourNick;
						setYourNick(fromName, true);
						setNick(toName, !isGroupChat); // ***
					}
					else {
// 						yourNick = toName; // ***
// 						nickName = fromName;
						setYourNick(toName, !isGroupChat); // ***
						setNick(fromName, !isGroupChat);
					}
// 					if (!isGroupChat) { // save the reliable (non-group chat) nicknames
// 						prevNick = nickName;
// 						prevYourNick = yourNick;
// 					}
					// User confirmed, we can add these to the known lists
					if (confirmedNew) {
						if (msgIsSend) {
							yourDisplayNames.add(fromName);
							userDisplayNames.add(toName);
						}
						else {
							userDisplayNames.add(fromName);
							if (!isGroupChat) yourDisplayNames.add(toName);
						}
					}
					return true;
				}
				
				s = reader.readToNext(">");
			}

			if (messageOpen) return true; // not terminated correctly. write it anyway.
			else {
				empty = true;
				reader.close();
				return false;
			}
		}
		
		/** MSNLog.getMsg()
		 */
		protected String getMsg() {
			return getTrMsg();
		}

		/** MSNLog.getSession()
		 */
		public String getSession() {
			// ugly workaround to print session nickname (sessNick)
			// because getTrSession() always uses nickName... fix this later
			String tmp = nickName;
			setNick(sessNick);
			String s = getTrSession();
			setNick(tmp);
			return s;
		}

	} //end class MSNLog


	/***
	 * MSNtoXML class, subclass of MSNLog
	 * Represents the XML version of the MSN log file
	 */
	static class MSNtoXML extends MSNLog {
		/** MSNtoXML() contructor
		 */
		public MSNtoXML(File f) throws IOException {
			super(f);
			priority = 2;
		}
		
		/** MSNtoXML.getMsg()
		 */
		protected String getMsg() {
			return getXMLMsg();
		}

		/** MSNtoXML.getSession()
		 */
		public String getSession() {
			// ugly workaround to print session nickname (sessNick)
			// because getTrSession() always uses nickName... fix this later
			String tmp = nickName;
			setNick(sessNick);
			String s = getXMLSession();
			setNick(tmp);
			return s;
		}
	}


	/**
	 * unHTMLEntity ( String )
	 *
	 * Returns a new string with various HTML escape sequences (e.g. &amp;) removed
	 * and replaced with the normal characters.
	 * 
	 * This one is specific for ICQ History Reader.
	 */
	static String unHTMLEntity(String s) {
		return s.replaceAll("&nbsp;"," ").replaceAll("&lt;","<").replaceAll("&gt;",">").replaceAll("&amp;","&");
	}
	
	/**
	 * unHTMLEntity2 ( String )
	 *
	 * Returns a new string with various XML escape sequences (e.g. &amp;) removed
	 * and replaced with the normal characters.
	 * 
	 * This one is specific for ICQ Lite messages, and MSN Messenger display names.
	 */
	static String unHTMLEntity2(String s) {
		return s.replaceAll("&lt;","<").replaceAll("&gt;",">").replaceAll("&quot;","\"").replaceAll("&amp;","&");
	}
	
	/**
	 * unHTMLEntityMSN ( String )
	 *
	 * Returns a new string with various XML escape sequences (e.g. &amp;) removed
	 * and replaced with the normal characters.
	 * 
	 * This one is specific for MSN Messenger messages (text).
	 */
	static String unHTMLEntityMSN(String s) {
		return s.replaceAll("&lt;","<").replaceAll("&gt;",">").replaceAll("&amp;","&");
	}

	/**
	 * numericEntity ( String )
	 *
	 * Returns a new string with all characters not in the set [0-9A-Za-z_] first 
	 * converted to their UTF-8 byte sequence, and then each byte replaced by 3 
	 * characters, %hh, where hh is the byte value in hex.
	 *
	 * Characters in lower ASCII (0-127) will become one %hh entity, whereas others 
	 * will become multiple entities, depending on the length of their UTF-8 encodings.
	 *
	 * This function is similar to rawurlencode() in PHP, except for .- and UTF-8.
	 */
	static String numericEntity(String s) {
		// make sure we escape unicode characters correctly - 
		// the way it is done, is to encode in UTF-8, then escape each byte in that encoding with %hh notation
		ByteArrayOutputStream byteStream;
		OutputStreamWriter sw;
		byte[] bytes;
		byte b;
		char c;
		String hh;
		
		byteStream = new ByteArrayOutputStream();
		try {
			sw = new OutputStreamWriter(byteStream, "utf8");
			sw.write(s);
			sw.close();
		} catch (UnsupportedEncodingException e) {
			System.out.println("Error: utf8 encoding not supported.");
		} catch (IOException e) {
			e.printStackTrace();
		}
		bytes = byteStream.toByteArray();

		// String out = "";
		StringBuffer out = new StringBuffer();

// 		for (int i=0; i<s.length(); i++) {
// 			c = s.charAt(i);
		for (int i=0; i<bytes.length; i++) {
			b = bytes[i];
			if ( ('0'<=b && b<='9') || ('a'<=b && b<='z') || ('A'<=b && b<='Z') || b=='_' ) {
				out.append((char)b);
			}
			else {
				hh = Integer.toHexString(0xff & b).toUpperCase();
				if (hh.length() < 2) hh = "0"+hh;
				else if (hh.length() > 2) System.out.println("Warning: Character did not become proper entity " + (char)b + "=" + hh);
				out.append("%"+hh);
			}
		}
		return out.toString();
	}
	
	/**
	 * trXMLEntity ( String )
	 *
	 * Encodes a string for use in Trillian XML logs. First encodes HTML entities, then encodes all 
	 * non-alphanumeric characters with %hh numeric (hex) entity.
	 */
	static String trXMLEntity(String s) {
		// in the XML, <a href (or encoded, %3Ca%20href) starts a link in the history.
		// if chat uses < or &, must be html entitified to prevent unintended behaviour
		// > is left as %3E (Trillian 3.1 behaviour)
		return numericEntity(s.replaceAll("&","&amp;").replaceAll("<","&lt;")); //.replaceAll(">","&gt;")
	}

	/**
	 * Longest common substring, with penalty for matches further away from the beginning.
	 * 
	 * Returns the similarity score as a float, where 0.0 is no match, and higher 
	 * scores mean greater similarity.
	 */
	static float lcsP(String s, String t) {
		//Uncomment all lines with len and res to find the actual string matches.
		int m = s.length();
		int n = t.length();
		//int[][] len = new int[m+1][n+1];
		float[][] score = new float[m+1][n+1];
		float best = 0;
		float cm = 1; // bonus for matching 1 letter
		float cd = -0.5F; // cost for being 1 letter away from front, per string
		//List<String> res = new Vector<String>();
		
		s = s.toLowerCase();
		t = t.toLowerCase();
		
		for (int i=0; i<=m; i++) {
			score[i][0] = i * cd;
		}
		for (int j=0; j<=n; j++) {
			score[0][j] = j * cd;
		}
		
		for (int i=1; i<=m; i++) {
			for (int j=1; j<=n; j++) {
				if (s.charAt(i-1) == t.charAt(j-1)) {
					score[i][j] = score[i-1][j-1] + cm;
					//len[i][j] = len[i-1][j-1] + 1;
					if (score[i][j] > best) {
						best = score[i][j];
						//res = new Vector<String>();
					}
					//if (score[i][j] == best) {
					//	res.add(s.substring(i-len[i][j],i));
					//}
				}
				else {
					score[i][j] = (i+j) * cd;
				}
			}
		}
				
		return best;
	}

	/**
	 * Hash function that MSN Messenger uses for its file names.
	 * Takes any string and hashes it to a integer from 0 to 2^32 - 1.
	 */
	static long msnHash(String s) {
		int hash = 0;
		char c;
		
		for (int i=0; i<s.length(); i++) {
			c = s.charAt(i);
			hash = 101*hash + c;
		}
		
		return hash & 0xffffffffL; // return the unsigned (positive) version of the 32-bit int
	}
	
	/**
	 * readIntFrom ( BufferedReader )
	 *
	 * Reads an integer from the stream, returns it.
	 * Will reprompt and retry the read until an integer is found.
	 * Throws IOException if an I/O error occurs.
	 */
	static int readIntFrom(BufferedReader reader) throws IOException{
		String utxt = reader.readLine();
		try {
			int i = Integer.parseInt(utxt.trim());
			return i;
		} catch (Exception e) {
			System.out.print("You must enter a number: ");
			return readIntFrom(reader);
		}
	}
	
	/**
	 * readYesNoFrom ( BufferedReader )
	 *
	 * Reads the stream for Yes or No, returning True or False, respectively.
	 * Will reprompt and retry the read until an answer is found.
	 * Throws IOException if an I/O error occurs.
	 */
	static boolean readYesNoFrom(BufferedReader reader) throws Exception{
		boolean yesno = false;
		String utxt = reader.readLine().trim() + " ";
		char first = utxt.charAt(0);
		if (first == 'Y' || first == 'y') return true;
		else if (first == 'N' || first == 'n') return false;
		else {
			System.out.print("Yes or No? ");
			return readYesNoFrom(reader);
		}
	}
	/*
	static void readTrBuddies (File buddiesFile, INIFile ini) {
		MyReader reader = new MyReader(new FileReader(buddiesFile));
		String s; // = reader.readLine();
		String str;
		StringTokenizer st;
		int startIdx, endIdx;

		while ( (s=reader.readLine()) != null ) {
			str = s.trim();
			if (str.startsWith("<buddy")) {
				startIdx = str.indexOf(">");
				endIdx = str.indexOf("</buddy>");
				if (startIdx == -1 || endIdx == -1) continue;
				String name = unNumericEntity(str.substring(startIdx, endIdx));

				int p = startIdx;
				startIdx = str.indexOf("uri=\"");
				endIdx = str.indexOf("\"", startIdx + 5);
				if (startIdx == -1 || startIdx > p || )
			}
		}
		
	}
*/

	/***
	 *  main ( String[] )
	 *
	 *  Asks the user for some information.
	 *  Then executes the merging of Trillian logs.
	 ***/
	public static void main(String[] args) throws Exception {
		
		System.out.println("IMmerge by zAlbee " + "(@" + "gmail.com)");
		System.out.println("v " + VERSION + " (" + LAST_MODIFY + ")");
		System.out.println("================================\n");
		
		Log.sdfMSN.setTimeZone(TimeZone.getTimeZone("GMT"));
		
		for (int i=0; i<args.length; i++) {
			String arg = args[i];
			if (arg.toLowerCase().equals("-defaultcharset")) {
				if (Charset.isSupported(args[i+1])) {
					defaultCharsetName = args[i+1];
					System.out.println("Will read using " + Charset.forName(defaultCharsetName).name() + " if UTF-8 not detected.");
				}
				else {
					System.out.println("Sorry, charset " + args[i+1] + " is not supported by your Java virtual machine.");
					System.out.println("Will read using your system default charset if UTF-8 not detected.");
				}
				i++;
			}
		}
		
		int numTrLog, numHtml, numICQLite, numMSN = 0;
		int yourUIN = 0;
		String yourNick = null;
		String yourMSNID = null;
		String yourMSNNick = null;

		String[] mediums = {"AIM", "ICQ", "IRC", "MSN", "YAHOO", "ASTRA", "FACEBOOK", "GOOGLE", "JABBER", "MYSPACE", "XFIRE"};
		String outPath = "output";

		File outDir = new File(outPath);
		if (!outDir.isDirectory() && !outDir.mkdir()) {
			System.out.println("Couldn't create '" + outPath + "' directory");
			System.out.println("Failed.");
			return;
		}
		
		settingsINI = new INIFile("immerge.ini");

		in = new BufferedReader(new InputStreamReader(System.in));
		System.out.print("How many folders (logs1 .. logsN) of Trillian-style <username>.log: ");
		numTrLog = readIntFrom(in);
		System.out.print("How many folders (icqh1 .. icqhN) of IcqHR output <UIN>.html (UIN per file): ");
		numHtml = readIntFrom(in);
		System.out.print("How many folders (icql1 .. icqlN) of ICQ Lite history: ");
		numICQLite = readIntFrom(in);
		System.out.print("How many folders (msn1 .. msnN) of MSN (Windows Live) history: ");
		numMSN = readIntFrom(in);

		if (numHtml > 0 || numICQLite > 0) {
			System.out.print("Your ICQ UIN/number (required): ");
			yourUIN = readIntFrom(in);
			System.out.print("Your ICQ Nickname (as you want it to appear): ");
			yourNick = in.readLine();
		}

		if (numHtml > 0) {
			System.out.print("Correct ICQHR timezone by how many hours? (-16 = default, 0 = no change): ");
			tz_adjust = readIntFrom(in);
		}
		
		if (numMSN > 0) {
			System.out.print("Your MSN login (e.g. abc@hotmail.com): ");
			yourMSNID = in.readLine().toLowerCase();
			System.out.print("Your MSN display name (most common or earliest): ");
			yourMSNNick = in.readLine();
			if (yourMSNNick.trim().length() == 0) yourMSNNick = yourMSNID;

			System.out.println("\nHow should unknown MSN display names be resolved?");
			System.out.println("  1. Automatically (no prompts, can be inaccurate)");
			System.out.println("  2. Prompt on vague names (<1.0 certainty) *recommended");
			System.out.println("  3. Prompt on any name that is new/uncertain");
			int msnPrompt = 0;
			while (msnPrompt < 1 || msnPrompt > 3) {
				System.out.print("Select (1-3): ");
				msnPrompt = readIntFrom(in);
			}
			promptOnNewName = msnPrompt == 3;
			promptOnVagueName = msnPrompt == 2;
		}

		if (numHtml > 0 || numICQLite > 0 || numMSN > 0) {
			System.out.print("Display time in 24-hr format? (yes = 24-hr, no = AM/PM): ");
			boolean use24h = readYesNoFrom(in);
			if (use24h) {
				Log.sdfTTime = Log.sdfTTime24;
			}
		}
		
		
		System.out.print("Should the logs be output to a \"Query\" or \"Channel\" folder like Trillian 3? (Yes/No): ");
		boolean useQuery = readYesNoFrom(in);
		
		System.out.println();
		
		// Loop through mediums, do find/merge for each

		for (int j=0; j<mediums.length; j++) {
			System.out.println("== " + mediums[j] + " ==");

			// link the username to a Vector of the logs
			Hashtable userLogs = new Hashtable();

			// link the username to a Vector of the XML files
			Hashtable userXMLs = new Hashtable();

			// Find files - Trillian logs
			for (int i=0; i<numTrLog; i++) {
				String logsPath = "logs" + (i+1) + File.separator + mediums[j]; // in folder
				String queryPath = logsPath + File.separator + "Query";         // in Query folder
				String channelPath = logsPath + File.separator + "Channel";     // in Channel folder
				
				File dir = new File(logsPath);
				File dirQ = new File(queryPath);
				File dirC = new File(channelPath);
				if (!dir.isDirectory()) continue;

				File[][] files = new File[3][];
				files[0] = dir.listFiles();
				files[1] = dirQ.isDirectory() ? dirQ.listFiles() : new File[0];
				files[2] = dirC.isDirectory() ? dirC.listFiles() : new File[0];

				for (int qd = 0; qd <= 2; qd++) {
					for (int k=0; k<files[qd].length; k++) {
						File f = files[qd][k];
						String filename = f.getName();

						// Trillian .log
						if (f.isFile() && filename.endsWith(".log")) {
							TrLog lg = new TrLog( f );
							String userCase = filename.substring(0, filename.length()-4);
							String user = userCase.toLowerCase();
							lg.trillID = userCase;
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}
						}

						// Trillian .xml
						if (qd!=0 && f.isFile() && filename.endsWith(".xml")) {
							
							if (filename.endsWith("-assets.xml")) {
								// assets, don't process - Trillian finds automatically
							}
							else {
								TrXML xml = new TrXML( f );
								String userCase = filename.substring(0, filename.length()-4);
								String user = userCase.toLowerCase();
								xml.trillID = userCase;
								if (userXMLs.containsKey(user))
									((Vector)userXMLs.get(user)).add(xml);
								else {
									Vector XMLs = new Vector();
									XMLs.add(xml);
									userXMLs.put(user, XMLs);
								}
							}
							
						}
						
					} // k (files)
				} // qd
				
			} // i (numTrLog)
		
			// Find files - ICQ History Reader 1.8 .html output
			if (j==ICQ) { // ICQ

				for (int i=0; i<numHtml; i++) {
					File dir = new File("icqh" + (i+1));
					if (!dir.isDirectory()) continue;

					File[] files;
					files = dir.listFiles();
					for (int k=0; k<files.length; k++) {
						String filename = files[k].getName().toLowerCase();

						if (files[k].isFile() && filename.endsWith(".html")) {
							String userCase = filename.substring(0, filename.length()-5); 
							String user = userCase.toLowerCase();
							ICQHRLog lg = new ICQHRLog( files[k] );
							lg.trillID = userCase;
							lg.setUserID(userCase);
							lg.setYourUserID(""+yourUIN);
							lg.setYourNick(yourNick);
							lg.init();
							
							// See if the init() function found a nickname, if so, save it.
							String nick = lg.getNick();
							if (!nick.equals(userCase)) {
								settingsINI.setStringProperty(INI_SECTION_ICQNAMES, user, nick, null);
							}
							
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}

							ICQHRtoXML xml = new ICQHRtoXML( files[k] );
							xml.trillID = userCase;
							xml.setUserID(userCase);
							xml.setYourUserID(""+yourUIN);
							xml.setYourNick(yourNick);
							xml.init();
							if (userXMLs.containsKey(user))
								((Vector)userXMLs.get(user)).add(xml);
							else {
								Vector XMLs = new Vector();
								XMLs.add(xml);
								userXMLs.put(user, XMLs);
							}

						}
					} // k (files)

				} // i (numHtml)
			} // if j==1 (icq)
			
			// Find files - ICQ Lite logs
			if (j==ICQ) { // ICQ

				for (int i=0; i<numICQLite; i++) {
					File dir = new File("icql" + (i+1));
					if (!dir.isDirectory()) continue;

					// inside should be many directories
					File[] subDirs;	
					subDirs = dir.listFiles();
					for (int k=0; k<subDirs.length; k++) {
						File subDir = subDirs[k];
						if (!subDir.isDirectory()) continue;
						String userCase = subDir.getName(); // their ICQ number
						String user = userCase.toLowerCase();
						
						// get nickname from INI file
						String nickName = settingsINI.getStringProperty(INI_SECTION_ICQNAMES, user);
						if (user.equals("unreadmsg")) {
							nickName = userCase;
						}
						else if (nickName == null) {
							System.out.print("Enter the nickname for user " + userCase + ": ");
							nickName = in.readLine();
							if (nickName.length() > 0) {
								settingsINI.setStringProperty(INI_SECTION_ICQNAMES, user, nickName, null);
							}
							else {
								nickName = userCase;
							}
						}
						
						File[] files = subDir.listFiles();
						
						for (int m=0; m<files.length; m++) {
							File file = files[m];
							String filename = file.getName();

							if (file.isFile() && filename.endsWith(".xml")) {
								ICQLiteLog lg = new ICQLiteLog( file );
								lg.trillID = userCase;
								lg.setUserID(userCase);
								lg.setNick(nickName);
								lg.setYourUserID(""+yourUIN);
								lg.setYourNick(yourNick);

								if (userLogs.containsKey(user))
									((Vector)userLogs.get(user)).add(lg);
								else {
									Vector logs = new Vector();
									logs.add(lg);
									userLogs.put(user, logs);
								}
	
								ICQLitetoXML xml = new ICQLitetoXML( file );
								xml.trillID = userCase;
								xml.setUserID(userCase);
								xml.setNick(nickName);
								xml.setYourUserID(""+yourUIN);
								xml.setYourNick(yourNick);
								
								if (userXMLs.containsKey(user))
									((Vector)userXMLs.get(user)).add(xml);
								else {
									Vector XMLs = new Vector();
									XMLs.add(xml);
									userXMLs.put(user, XMLs);
								}
	
							}
						} // m (files)
					} // k (subDirs)

				} // i (numICQLite)
			} // if j==ICQ
			
			// Find files - MSN (Windows Live) logs
			if (j==MSN) {
				for (int i=0; i<numMSN; i++) {
					File dir = new File("msn" + (i+1));
					if (!dir.isDirectory()) continue;

					File[] files;
					files = dir.listFiles();
					for (int k=0; k<files.length; k++) {
						String filename = files[k].getName();

						if (files[k].isFile() && filename.endsWith(".xml")) {
							String filenameNoExt = filename.substring(0, filename.length()-4);
							// Attempt to find user ID from hash.
							String user = settingsINI.getStringProperty(INI_SECTION_MSNUSERS, filenameNoExt);

							// generally hash is 9 or 10 digits, sometimes 8
							for (int hlen=8; hlen<=10 && user==null; hlen++) {
								int brk = filenameNoExt.length() - hlen;
								try {
									long hash = Long.parseLong(filenameNoExt.substring(brk));
									String user1 = filenameNoExt.substring(0, brk);
									if (msnHash(user = user1 + "@hotmail.com") == hash) {
										break; // success!
									}
									else if (msnHash(user = user1 + "@msn.com") == hash) {
										break; // success!
									}
									else if (msnHash(user = user1 + "@gmail.com") == hash) {
										break; // success!
									}
									else {
										user = null;
									}
								}
								catch (Exception e) {
								}
							}
							
							boolean skip = false;
							while (user == null) {
								System.out.print("Enter the MSN ID (full email address) for " + filenameNoExt + " or type 'skip': ");
								user = in.readLine();
								if (user.toLowerCase().equals("skip")) {
									skip = true;
									break;
								}
								if (user.indexOf('@') == -1) {
									user = null;
									continue;
								}
								String test = user.substring(0, user.indexOf('@')) + msnHash(user);
								if ( !test.equals(filenameNoExt) ) {
									user = null;
									System.out.println("Hash does not match (" + test + ")");
								}
							}

							if (skip) {
								System.out.println("File " + filename + " will not be merged.");
								continue; // go to next file
							}
							
							settingsINI.setStringProperty(INI_SECTION_MSNUSERS, filenameNoExt, user, null);

							MSNLog lg = new MSNLog( files[k] );
							lg.trillID = user;
							lg.setNick(user);
							lg.setYourNick(yourMSNNick);
							lg.setUserID(user);
							lg.setYourUserID(yourMSNID);
														
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}

							MSNtoXML xml = new MSNtoXML( files[k] );
							xml.trillID = user;
							xml.setNick(user);
							xml.setYourNick(yourMSNNick);
							xml.setUserID(user);
							xml.setYourUserID(yourMSNID);

							if (userXMLs.containsKey(user))
								((Vector)userXMLs.get(user)).add(xml);
							else {
								Vector XMLs = new Vector();
								XMLs.add(xml);
								userXMLs.put(user, XMLs);
							}

						}
					} // k (files)

				} // i (numMSN)
				
			} // if j==MSN


			String outPath2 = outPath + File.separator + mediums[j];
			String outPathQuery = outPath2 + File.separator + "Query";
			String outPathChannel = outPath2 + File.separator + "Channel";


			/** ---- merge Logs ---- **/
			
			// if logs exist, make output dir
			if (userLogs.size() > 0) {
				System.out.println("Starting merge, output format: Trillian Log\n");

				//outPath2[j] = outPath + File.separator + mediums[j] + (useQuery? File.separator+"Query" : "");
				
				String outPath3 = outPathQuery;
				String outPath3Ch = outPathChannel;
				
				if (!useQuery) {
					File outDir2 = new File(outPath2);
					if (!outDir2.isDirectory() && !outDir2.mkdirs()) {
						System.out.println("Error: Couldn't create directory '" + outPath2 + "'");
						System.out.println("Skipping " + mediums[j]);
						continue;
					}
					outPath3 = outPath3Ch = outPath2;
				}
				else {
					File outDir3 = new File(outPathQuery);
					if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
						System.out.println("Error: Couldn't create directory '" + outPathQuery + "'");
						System.out.println("Skipping " + mediums[j]);
						continue;
					}
					outDir3 = new File(outPathChannel);
					if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
						System.out.println("Error: Couldn't create directory '" + outPathChannel + "'");
						System.out.println("Skipping " + mediums[j]);
						continue;
					}
				}

				// Iterate through the logs.
				int users = 0;
				Enumeration userEnum = userLogs.keys();
				Vector userNames = new Vector();
	
				while (userEnum.hasMoreElements()) {
					users++;
					userNames.add((String)userEnum.nextElement());
				}
				Collections.sort(userNames);
	
				for (int u=0; u<users; u++) {
					int sessions = 0;
					int dupes = 0;
					int numFiles;
	
					String user = (String)userNames.elementAt(u);
					Vector logs = (Vector)userLogs.get(user);
	
					// Initialize; grab one session from each log, and record in log buffer
					for (int i=0; i<logs.size(); i++) {
						Log lg = (Log)logs.get(i);
						lg.init();
						//lg.sessDate = new Date(i); // initialize date, in case the readSession() func doesn't find a session date.
						lg.sessDate = new Date(-1000); // initialize date, in case the readSession() func doesn't find a session date.
						if (!lg.readSession()) logs.remove(i--);
					}
					Collections.sort(logs);
					numFiles = logs.size();
					
					// Get case-sensitive user ID (v0.63)
					if (numFiles>0) user = ((Log)logs.get(0)).trillID;
					
					boolean useChDir = (j==IRC && user.indexOf("#") != -1) || (j==MSN && user.startsWith("Group Conversation"));
	
					File f = new File( ( useChDir ? outPath3Ch : outPath3 ) + File.separator + user + ".log");
					if (!f.exists()) f.createNewFile();
	
					try {
						OutputStreamWriter fw = new OutputStreamWriter(new FileOutputStream(f), "utf8");
						fw.write(UNICODEBOM);
						
						while (logs.size()>0) {
							sessions++;
							Log lg = (Log)logs.get(0);
							fw.write(lg.getSession());
	
							for (int i=1; i<logs.size(); i++) {
								Log olog = (Log)logs.get(i);
								if (!lg.sessDate.equals(olog.sessDate)) {
									//System.out.println("No Dup - " + lg.sessDate + " vs " + olog.sessDate);
									break;
								}
	
								dupes++;
								//System.out.println("-- Duplicate in " + lg.filename + "(" + lg.pathname + " & " + olog.pathname + ") on " + lg.sessDate);
								if (!olog.readSession()) logs.remove(i--);
							}
	
							if (!lg.readSession()) logs.remove(0);
							Collections.sort(logs); // 0.62
						}
	
						fw.close();
	
						System.out.println(user + ": merged " + numFiles + " files, " + sessions + " sessions, " + dupes + " duplicates");
					}
					catch (IOException e) {
						e.printStackTrace();
					}
	
				} //end for users
	
				System.out.println(users + " Log files written.\n");
			}


			/** ---- merge XML ---- **/

			// if XML files exist, make output dir
			if (userXMLs.size() > 0) {
				System.out.println("Starting merge, output format: Trillian XML\n");
				
				File outDir3 = new File(outPathQuery);
				if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
					System.out.println("Error: Couldn't create directory '" + outPathQuery + "'");
					System.out.println("Skipping " + mediums[j]);
					continue;
				}
				outDir3 = new File(outPathChannel);
				if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
					System.out.println("Error: Couldn't create directory '" + outPathChannel + "'");
					System.out.println("Skipping " + mediums[j]);
					continue;
				}

				// Iterate through the XML files.
				int users = 0;
				Enumeration userEnum = userXMLs.keys();
				Vector userNames = new Vector();
	
				while (userEnum.hasMoreElements()) {
					users++;
					userNames.add((String)userEnum.nextElement());
				}
				Collections.sort(userNames);
	
				for (int u=0; u<users; u++) {
					int sessions = 0;
					int dupes = 0;
					int numFiles;
					
					int numMinus = 0;
					int numMinusDupes = 0;
					Date minusDate = new Date(-1000);
	
					String user = (String)userNames.elementAt(u);
					Vector XMLs = (Vector)userXMLs.get(user);
					
					debug("xml: " + user);
										
					// Initialize; grab one session from each log, and record in log buffer
					for (int i=0; i<XMLs.size(); i++) {
						Log xml = (Log)XMLs.get(i);
						debug("xmls[" + i + "].init ");
						xml.init();
						debug("setting sess date = " + (i - XMLs.size() - 1000));
						//xml.sessDate = new Date( i - XMLs.size() - 1000); // initialize date, in case the readSession() func doesn't find a session date.
						xml.sessDate = minusDate; // initialize date, in case the readSession() func doesn't find a session date.
						if (!xml.readSession()) XMLs.remove(i--);
						debug("session read");
					}
					Collections.sort(XMLs);
					numFiles = XMLs.size();

					// Get case-sensitive user ID (v0.63)
					if (numFiles>0) user = ((Log)XMLs.get(0)).trillID;
					
					boolean useChDir = (j==IRC && user.indexOf("#") != -1) || (j==MSN && user.startsWith("Group Conversation"));

					File f = new File( ( useChDir ? outPathChannel : outPathQuery ) + File.separator + user + ".xml");
					if (!f.exists()) f.createNewFile();
	
					try {
						FileWriter fw = new FileWriter(f);
						//Trillian doesn't use UTF-8 for the XML files, but is it safe to use system default encoding?

						while (XMLs.size()>0) {
							sessions++;
							Log xml = (Log)XMLs.get(0);
							fw.write(xml.getSession());
	
							if (xml.sessDate.equals(minusDate))
								numMinus++;
	
							for (int i=1; i<XMLs.size(); i++) {
								Log oxml = (Log)XMLs.get(i);
								if (!xml.sessDate.equals(oxml.sessDate) ) {
									//System.out.println("No Dup - " + xml.sessDate + " vs " + oxml.sessDate);
									break;
								}
	
								if (xml.sessDate.equals(minusDate)) {
									// if we find two sessions with -1, we can't be sure that they are equal, but we will treat them that way
									numMinusDupes++;
								}
								dupes++;
								//System.out.println("-- Duplicate in " + xml.filename + "(" + xml.pathname + " & " + oxml.pathname + ") on " + xml.sessDate);
								if (!oxml.readSession()) XMLs.remove(i--);
							}
	
							if (!xml.readSession()) XMLs.remove(0);
							Collections.sort(XMLs); // 0.62
						}
	
						fw.close();
	
						System.out.println(user + ": merged " + numFiles + " XML files, " + sessions + " sessions, " + dupes + " duplicates");
						if (numMinus > 0)
							System.out.println(" * Warning: kept " + numMinus + " sessions with a -1 date, an additional " + numMinusDupes + " may have been incorrectly dropped as duplicates");
					}
					catch (IOException e) {
						e.printStackTrace();
					}

				} //end while enumeration of users
	
				System.out.println(users + " XML files written.\n");
			}

			else {
				System.out.println("No files, skipping...\n");
				continue;
			}



		} //end for mediums

		in.close();

		System.out.println("Saving settings to " + settingsINI.getFileName());
		settingsINI.save();

		System.out.println("Finished.");

	} // end main

}
